#!/opt/anaconda1anaconda2anaconda3/bin/perl
use Codon;			# package for Codon calculations
				# Make sure Codon.pm is where this file
				# can see it; in particular, keep Codon.pm
				# in the same directory as this file
## Covariances for all pairs of pairwise sequence comparisons
## Takes some run time; 21 gag sequneces took 2.8 minutes on our linux box,
## to do the ~11 million calculations required.  Scales to the 4th power
## in terms of numbers of sequences added, linearly in terms of number of codons. 
## so may take a quite a while for large sequence sets.
## Based on the statistic described in: T. Ota and M. Nei, Variance and Covariances of the
## Numbers of Synonymous and Nonsynonymous Substitutions per Site,
## Mol. Biol. Evol. 11(4):613-619, 1994.
## See also: M. Nei and L. Jin, Variances of the Average Numbers
## of Nucleotide Substitutions Within and Between Populations,
## Mol. Biol. Evol. 6(3):290-300, 1989.

## Takes as input a file codons.$$, where $$ is a number; this
## is one of the output files from 'SNAP.pl'

#SNAPstats.pl (Synonymous Nonsynonymous Analysis Program)
#Written by: James Theiler and Bette Korber, T10 MS K710, LANL, Los Alamos, NM 87545
#(505) 665-4453, btk@t10.lanl.gov

# Copyright, 1998. The regents of the University of California.  This
# software was produced under a US Government DOE-NIH contract at Los
# Alamos National Laboratory, is operated by the University of
# California for the US Department of Energy (DOE).  The US government
# is licensed to use, reproduce, and distribute this software.
# Permission is granted to the public to copy and use this software
# without charge, provided that this Notice and any statement of
# authorship are reproduced on all copies, and this software
# is not sold. Neither the government nor
# the university makes any warranty, express or implied, or assumes any
# liability or responsibility for the use of this software.

($me = $0) =~ s,.*/,,;

sub usage {
    print <<EndUsage;
Usage: $me codons.\#\#
       Computes covariances for all pairwise comparisons.

Actually, computes covariances for all _pairs_ of pairwise comparisons!

EndUsage
    die "\n";
}

## "my" means variable is local to its "block"
## In this case, the local block is the whole file
my $codonInputFile = shift @ARGV or &usage();

open(CIF,"$codonInputFile") || 
    die "Cant open codon input file [$codonInputFile]\n";

my @pairList=();
my $pair = "";
while (<CIF>) {
    chomp;
    next if /^Codon/;
    if (/This is comparison (\d+)\s*x\s*(\d+)/) {
	if ($pair) {
	    #print STDERR "S_$pair = ",$sumSynPoten{$pair},"\n";
	}
	$pair = "$1,$2";
	push @pairList,$pair;
	$sp = $synPoten{$pair} = [];
	$s = $syn{$pair} = [];
	$sumSynPoten{$pair} = 0;
	$np = $nonPoten{$pair} = [];
	$n = $non{$pair} = [];
	$sumNonPoten{$pair} = 0;
	$nCodons{$pair} = 0;
	next;
    }
    ($i,$class,$ca,$cb,$aaa,$aab,$syn,$non) = split;
    if ($ca =~ /[-N]/ || $cb =~ /[-N]/) {
	$syn=$spi=0;
	$non=$npi=0;
    } else {
	## Codon::to_synPotential() calls function to_synPotential
	## in package Codon
	$spi = $sp->[$i] = (Codon::to_synPotential($ca) 
			    + Codon::to_synPotential($cb))/2;
	$npi = 3-$spi;
    }
    $syn ||= 0;
    $non ||= 0;
    $s->[$i] = $syn;
    $n->[$i] = $non;
    $sp->[$i]= $spi;
    $np->[$i]= $npi;
    $sumSynPoten{$pair} += $spi;
    $sumNonPoten{$pair} += $npi;
    $nCodons{$pair} += 1;
}
close CIF;

my $pSyn_ave=0;
my $pNon_ave=0;

foreach $pair (@pairList) {
    my $s = $syn{$pair};
    my $sp = $synPoten{$pair};
    my $n = $non{$pair};
    my $np = $nonPoten{$pair};
    my $ssum = 0;
    my $nsum = 0;
    for ($i=0; $i< $nCodons{$pair}; ++$i) {
	$ssum += $s->[$i];
	$nsum += $n->[$i];
    }
    my $ps = $ssum/$sumSynPoten{$pair};
    my $pn = $nsum/$sumNonPoten{$pair};
    ## Replacing sdmni with sdmni-psmn*smni, in the
    ## notation of Ota and Nei, eq 13.
    for ($i=0; $i< $nCodons{$pair}; ++$i) {
	$s->[$i] -= $ps*($sp->[$i]);
	$n->[$i] -= $pn*($np->[$i]);
    }
    $pSyn_ave += $ps;
    $pNon_ave += $pn;

    $psTable{$pair} = $ps;
    $pnTable{$pair} = $pn;

    $dsTable{$pair} = -0.75*log(1-4*$ps/3);
    $dSyn_ave += $dsTable{$pair};

    $dnTable{$pair} = -0.75*log(1-4*$pn/3);
    $dNon_ave += $dnTable{$pair};

    #print STDERR "Pair: $pair, ss=$ssum, sn=$nsum, ps=$ps, pn=$pn\n";
}

my $nPairs = scalar(@pairList);

$pSyn_ave /= $nPairs;
$pNon_ave /= $nPairs;
$dSyn_ave /= $nPairs;
$dNon_ave /= $nPairs;

    
my $sum_scov=0;
my $sum_ncov=0;
my $sum_ps=0;
my $sum_pn=0;

for ($ipa=0; $ipa<$nPairs; ++$ipa) {
    $pa = $pairList[$ipa];
    for ($ipb=$ipa; $ipb<$nPairs; ++$ipb) {
	$pb = $pairList[$ipb];
	if ($nCodons{$pa} != $nCodons{$pb}) {
	    warn "Pair $pa and Pair $pb have unequal number of codons\n";
	}
	$sa = $syn{$pa};
	$sb = $syn{$pb};
	$na = $non{$pa};
	$nb = $non{$pb};
	for ($i=0; $i< $nCodons{$pa}; ++$i) {
	    $scov += ($sa->[$i])*($sb->[$i]);
	    $ncov += ($na->[$i])*($nb->[$i]);
	}
	$scov /= ($sumSynPoten{$pa})*($sumSynPoten{$pb});
	$ncov /= ($sumNonPoten{$pa})*($sumNonPoten{$pb});
	if ($ipa != $ipb) {
	    $scov *= 2;
	    $ncov *= 2;
	}
	$sum_scov += $scov;
	$sum_ncov += $ncov;

	$pd_correction = (1-4*$psTable{$pa}/3)*(1-4*$psTable{$pb}/3);
	$sum_scov_d += $scov/$pd_correction;
	$sum_ncov_d += $ncov/$pd_correction;
	

    }
}
## Divide sum_scov by nPairs^2, following Nei and Jin, eq (3).
print "\n";
printf "Average ps:         %.6f\n",$pSyn_ave;
printf "Variance(ps):       %.6f\n",$sum_scov/($nPairs*$nPairs);
printf "Std Deviation(ps):  %.6f\n",sqrt($sum_scov)/$nPairs;
printf "Average ds:         %.6f\n",$dSyn_ave;
printf "Variance(ds):       %.6f\n",$sum_scov_d/($nPairs*$nPairs);
printf "Std Deviation(ds):  %.6f\n",sqrt($sum_scov_d)/$nPairs;
printf "\n";
printf "Average pn:         %.6f\n",$pNon_ave;
printf "Variance(pn):       %.6f\n",$sum_ncov/($nPairs*$nPairs);
printf "Std Deviation(pn):  %.6f\n",sqrt($sum_ncov)/$nPairs;
printf "Average dn:         %.6f\n",$dNon_ave;
printf "Variance(dn):       %.6f\n",$sum_ncov_d/($nPairs*$nPairs);
printf "Std Deviation(dn):  %.6f\n",sqrt($sum_ncov_d)/$nPairs;




__END__
